@@ -1,7 +1,8 @@
 use std;
 
-use combinator::{AndThen, Assert, Buffered, Collect, CollectN, DecoderChain, Length, Map, MapErr,
-                 MaxBytes, MaybeEos, Omittable, SkipRemaining, Slice, TryMap};
+use combinator::{AndThen, Collect, CollectN, Length, Map, MapErr, MaxBytes, MaybeEos, Omittable,
+                 Peekable, Slice, TryMap};
+use tuple::TupleDecoder;
 use {ByteCount, Eos, Error, ErrorKind, Result};
 
 /// This trait allows for decoding items from a byte sequence incrementally.
@@ -9,17 +10,14 @@ pub trait Decode {
     /// The type of items to be decoded.
     type Item;
 
-    /// Consumes the given buffer (a part of a byte sequence), and decodes an item from it.
+    /// Consumes the given buffer (a part of a byte sequence), and proceeds the decoding process.
     ///
-    /// The first element of a succeeded result is the number of bytes consumed
-    /// from `buf` by the decoding process.
+    /// It returns the number of bytes consumed from the input buffer.
     ///
-    /// If an item is successfully decoded, the decoder will return `Ok((_, Some(..)))`.
-    ///
-    /// If the buffer does not contain enough bytes to decode the next item,
-    /// the decoder will return `Ok((_, None))`.
-    /// In this case, the decoder must consume as many bytes in the buffer as possible.
+    /// If an item is completely decoded, the next invocation of `is_idle` method will return `true`.
+    /// And if `is_idle` method returns `true`, `decode` method should consume no bytes.
     ///
+    /// The decoder must consume as many bytes in the buffer as possible.
     /// If an item is not yet decoded but the number of consumed bytes in the last `decode` invocation
     /// is smaller than the length of `buf`, it means the decoder has been suspended its work in any reasons.
     /// In that case the decoder may require some instructions from clients to resume the work,
@@ -27,7 +25,7 @@ pub trait Decode {
     ///
     /// # Errors
     ///
-    /// Decoders return the following kinds of errors as necessary:
+    /// The following errors may be returned by the decoder:
     /// - `ErrorKind::DecoderTerminated`:
     ///   - If all decodable items have been decoded,
     ///     the decoder must return this kind of error when `decode` method is called.
@@ -37,9 +35,28 @@ pub trait Decode {
     /// - `ErrorKind::InvalidInput`:
     ///   - Decoded items have invalid values
     ///   - Invalid parameters were given to decoders
-    /// - `ErrorKind::Error`:
+    /// - `ErrorKind::InconsistentState`:
+    ///   - The state of the decoder bocame inconsistent
+    ///   - This means the implementation contains a bug
+    /// - `ErrorKind::Other`:
     ///   - Other errors
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)>;
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize>;
+
+    /// Finishes the current decoding process and returns the decoded item.
+    ///
+    /// # Errors
+    ///
+    /// The following errors may be returned by the decoder:
+    /// - `ErrorKind::IncompleteDecoding`:
+    ///   - The decoding process has not been completed
+    /// - `ErrorKind::DecoderTerminated`:
+    ///   - The decoder has terminated (i.e., cannot decode any more items)
+    /// - `ErrorKind::InconsistentState`:
+    ///   - The state of the decoder bocame inconsistent
+    ///   - This means the implementation contains a bug
+    /// - `ErrorKind::Other`:
+    ///   - Other errors
+    fn finish_decoding(&mut self) -> Result<Self::Item>;
 
     /// Returns the lower bound of the number of bytes needed to decode the next item.
     ///
@@ -51,29 +68,57 @@ pub trait Decode {
     ///   - The next invocation of `decode()` will return it without consuming any bytes
     /// - (b) There are no decodable items
     ///   - All decodable items have been decoded, and the decoder cannot do any further works
+    ///   - In this case, the next invocation of `decode` method will fail.
     fn requiring_bytes(&self) -> ByteCount;
+
+    /// Returns `true` if there are no items to be decoded by the decoder
+    /// at the next invocation of `decode` method, otherwise `false`.
+    ///
+    /// Typically, `true` means the decoder already has a decoded item and
+    /// it is waiting for `finish_decoding` to be called.
+    ///
+    /// The default implementation returns the result of `self.requiring_bytes() == ByteCount::Finite(0)`.
+    fn is_idle(&self) -> bool {
+        self.requiring_bytes() == ByteCount::Finite(0)
+    }
 }
 impl<'a, D: ?Sized + Decode> Decode for &'a mut D {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         (**self).decode(buf, eos)
     }
 
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        (**self).finish_decoding()
+    }
+
     fn requiring_bytes(&self) -> ByteCount {
         (**self).requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        (**self).is_idle()
+    }
 }
 impl<D: ?Sized + Decode> Decode for Box<D> {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         (**self).decode(buf, eos)
     }
 
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        (**self).finish_decoding()
+    }
+
     fn requiring_bytes(&self) -> ByteCount {
         (**self).requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        (**self).is_idle()
+    }
 }
 
 /// An extension of `Decode` trait.
@@ -124,7 +169,7 @@ pub trait DecodeExt: Decode + Sized {
     /// assert_eq!(item, 8);
     /// # }
     /// ```
-    fn try_map<F, T, E>(self, f: F) -> TryMap<Self, F, T, E>
+    fn try_map<T, E, F>(self, f: F) -> TryMap<Self, T, E, F>
     where
         F: Fn(Self::Item) -> std::result::Result<T, E>,
         Error: From<E>,
@@ -159,14 +204,14 @@ pub trait DecodeExt: Decode + Sized {
     /// UnexpectedEos (cause; assertion failed: `!eos.is_reached()`; \
     ///                self.offset=1, self.bytes.as_ref().len()=2)
     /// HISTORY:
-    ///   [0] at src/bytes.rs:155
-    ///   [1] at src/fixnum.rs:192
+    ///   [0] at src/bytes.rs:153
+    ///   [1] at src/fixnum.rs:199
     ///   [2] at src/decode.rs:12 -- oops!
     ///   [3] at src/io.rs:44
     ///   [4] at src/decode.rs:16\n");
     /// # }
     /// ```
-    fn map_err<F, E>(self, f: F) -> MapErr<Self, F, E>
+    fn map_err<E, F>(self, f: F) -> MapErr<Self, E, F>
     where
         F: Fn(Error) -> E,
         Error: From<E>,
@@ -201,34 +246,6 @@ pub trait DecodeExt: Decode + Sized {
         AndThen::new(self, f)
     }
 
-    /// Takes two decoders and creates a new decoder that decodes both items in sequence.
-    ///
-    /// Chains are started by calling `StartDecoderChain::chain` method.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use bytecodec::{Decode, DecodeExt, StartDecoderChain};
-    /// use bytecodec::fixnum::U8Decoder;
-    /// use bytecodec::io::IoDecodeExt;
-    ///
-    /// let mut decoder = StartDecoderChain
-    ///     .chain(U8Decoder::new())
-    ///     .chain(U8Decoder::new())
-    ///     .chain(U8Decoder::new());
-    ///
-    /// let mut input = &b"foobar"[..];
-    ///
-    /// let item = decoder.decode_exact(&mut input).unwrap();
-    /// assert_eq!(item, (b'f', b'o', b'o'));
-    ///
-    /// let item = decoder.decode_exact(&mut input).unwrap();
-    /// assert_eq!(item, (b'b', b'a', b'r'));
-    /// ```
-    fn chain<D: Decode>(self, other: D) -> DecoderChain<Self, D, Self::Item> {
-        DecoderChain::new(self, other)
-    }
-
     /// Creates a decoder for collecting decoded items.
     ///
     /// # Examples
@@ -317,27 +334,6 @@ pub trait DecodeExt: Decode + Sized {
         Omittable::new(self, do_omit)
     }
 
-    /// Creates a decoder for skipping the remaining bytes in an input byte sequence
-    /// after decoding an item by using `self`.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// use bytecodec::{Decode, DecodeExt};
-    /// use bytecodec::fixnum::U8Decoder;
-    /// use bytecodec::io::IoDecodeExt;
-    ///
-    /// let mut input = &b"foo"[..];
-    ///
-    /// let mut decoder = U8Decoder::new().skip_remaining();
-    /// let item = decoder.decode_exact(&mut input).unwrap();
-    /// assert_eq!(item, b'f');
-    /// assert!(input.is_empty());
-    /// ```
-    fn skip_remaining(self) -> SkipRemaining<Self> {
-        SkipRemaining::new(self)
-    }
-
     /// Creates a decoder that will fail if the number of consumed bytes exceeds `bytes`.
     ///
     /// # Examples
@@ -359,29 +355,11 @@ pub trait DecodeExt: Decode + Sized {
         MaxBytes::new(self, bytes)
     }
 
-    /// Creates a decoder that will fail if the given assertion function returns `false`.
+    /// Takes two decoders and creates a new decoder that decodes both items in sequence.
     ///
-    /// # Examples
-    ///
-    /// ```
-    /// use bytecodec::{Decode, DecodeExt, ErrorKind};
-    /// use bytecodec::fixnum::U8Decoder;
-    /// use bytecodec::io::IoDecodeExt;
-    ///
-    /// let mut decoder = U8Decoder::new().assert(|&b| b == 3);
-    ///
-    /// let mut input = &[3, 4][..];
-    /// let item = decoder.decode_exact(&mut input).unwrap();
-    /// assert_eq!(item, 3);
-    ///
-    /// let error = decoder.decode_exact(&mut input).err();
-    /// assert_eq!(error.map(|e| *e.kind()), Some(ErrorKind::InvalidInput));
-    /// ```
-    fn assert<F>(self, f: F) -> Assert<Self, F>
-    where
-        F: for<'a> Fn(&'a Self::Item) -> bool,
-    {
-        Assert::new(self, f)
+    /// This is equivalent to call `TupleDecoder::new((self, other))`.
+    fn chain<T: Decode>(self, other: T) -> TupleDecoder<(Self, T)> {
+        TupleDecoder::new((self, other))
     }
 
     /// Creates a decoder that makes it possible to slice the input byte sequence in arbitrary units.
@@ -401,49 +379,44 @@ pub trait DecodeExt: Decode + Sized {
     /// let input = b"fboaor";
     /// let mut offset = 0;
     ///
-    /// let mut last_item0 = None;
-    /// let mut last_item1 = None;
     /// for _ in 0..3 {
     ///     decoder0.set_consumable_bytes(1);
-    ///     let (size, item) = decoder0.decode(&input[offset..], eos).unwrap();
-    ///     offset += size;
-    ///     last_item0 = item;
+    ///     offset += decoder0.decode(&input[offset..], eos).unwrap();
     ///
     ///     decoder1.set_consumable_bytes(1);
-    ///     let (size, item) = decoder1.decode(&input[offset..], eos).unwrap();
-    ///     offset += size;
-    ///     last_item1 = item;
+    ///     offset += decoder1.decode(&input[offset..], eos).unwrap();
     /// }
     ///
     /// assert_eq!(offset, input.len());
-    /// assert_eq!(last_item0, Some("foo".to_owned()));
-    /// assert_eq!(last_item1, Some("bar".to_owned()));
+    /// assert_eq!(decoder0.finish_decoding().unwrap(), "foo");
+    /// assert_eq!(decoder1.finish_decoding().unwrap(), "bar");
     /// ```
     fn slice(self) -> Slice<Self> {
         Slice::new(self)
     }
 
-    /// Creates a decoder that buffers the last decoded item.
+    /// Creates a decoder that enables to peek decoded items before calling `finish_decoding` method.
     ///
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeExt, Eos, StartDecoderChain};
+    /// use bytecodec::{Decode, DecodeExt, Eos};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::tuple::TupleDecoder;
     ///
-    /// let mut decoder = StartDecoderChain
-    ///     .chain(U8Decoder::new())
-    ///     .chain(U8Decoder::new())
-    ///     .chain(U8Decoder::new())
-    ///     .buffered();
-    /// let (size, item) = decoder.decode(b"foo", Eos::new(false)).unwrap();
+    /// let mut decoder = TupleDecoder::new((
+    ///     U8Decoder::new(),
+    ///     U8Decoder::new(),
+    ///     U8Decoder::new(),
+    /// )).peekable();
+    /// let size = decoder.decode(b"foo", Eos::new(false)).unwrap();
     /// assert_eq!(size, 3);
-    /// assert_eq!(item, None);
-    /// assert_eq!(decoder.take_item(), Some((b'f', b'o', b'o')));
-    /// assert_eq!(decoder.has_item(), false);
+    /// assert_eq!(decoder.peek(), Some(&(b'f', b'o', b'o')));
+    /// assert_eq!(decoder.finish_decoding().unwrap(), (b'f', b'o', b'o'));
+    /// assert_eq!(decoder.peek(), None);
     /// ```
-    fn buffered(self) -> Buffered<Self> {
-        Buffered::new(self)
+    fn peekable(self) -> Peekable<Self> {
+        Peekable::new(self)
     }
 
     /// Creates a decoder that ignores EOS if there is no item being decoded.
@@ -482,10 +455,9 @@ pub trait DecodeExt: Decode + Sized {
     /// );
     /// ```
     fn decode_from_bytes(&mut self, buf: &[u8]) -> Result<Self::Item> {
-        let (size, item) = track!(self.decode(buf, Eos::new(true)))?;
+        let size = track!(self.decode(buf, Eos::new(true)))?;
         track_assert_eq!(size, buf.len(), ErrorKind::InvalidInput);
-        let item = track_assert_some!(item, ErrorKind::InvalidInput);
-        Ok(item)
+        track!(self.finish_decoding())
     }
 }
 impl<T: Decode> DecodeExt for T {}
